##########################################################################
##
## File Name: 'SampleScript.R'
##
## Author:    FHRB Toledo < f.toledo@cgiar.org >
##
## Date:      Wed May 15, 2019
##
## Source:    Basic pipeline to play with isqg
##            supplemental material for the G3 paper
##
##            ... several -- read comments (# in R and // in C++)
##
###########################################################################

rm(list = objects()); ls() # CLEAR 'workspace'             (R  environment)

## NOTE for windows and mac users:
## Be sure you have installed the compilers and libraries
## Please follow the instruction on these links:
## windows: https://cran.r-project.org/bin/windows/Rtools/installer.html
## mac: https://cran.r-project.org/bin/macosx/tools/

## needed packages
packs <- c('isqg', 'reshape2', 'plyr', 'Rcpp', 'ggplot2')

## get non installed
toinstall <- names(which(!sapply(packs,
                                 function(pack) nzchar(system.file(package = pack)))))

## install non installed packages
if(length(toinstall) != 0)
    install.packages(pkgs = toinstall,
                     repos = "https://cloud.r-project.org",
                     dependencies = TRUE,
                     type = 'source',
                     verbose = TRUE)

## load libraries
sapply(packs, require, character.only = TRUE)

## load simple map
data(ToyMap) # example of input data
head(ToyMap, 10) # 2 chromosomes, 42 markers

## start a standard _in silico_ specie
## meiosis follow a count location process:
## ref: Karlin & Liberman (1978) [Proc. Natl. Acad. Sci. 75(12):6332--6336]
spp <- set_specie(ToyMap)

## start some 'founders' individuals from specie
AA <- spp$founder(code = 'AA')
aa <- spp$founder(code = 'aa')
## alternative to generate heterozigous individuals
Aa <- spp$founder(code = 'Aa')
aA <- spp$founder(code = 'aA')

## sampling a 'random' individual -- can be your real data
real <- sample(c('2 2', '2 1', '1 2', '1 1'),
               size = nrow(ToyMap), replace = TRUE)
names(real) <- ToyMap$snp

## start an individual from outside
Gidi <- import(spp, real)

## start an individual as the complete oposite from someone
Gidj <- Gidi$mirror()

## making some crosses (cross, selfcross and dh)
## are accessible by standalone function and methods
F1 <- cross(n = 1, AA, aa)  # the hybrid
F2 <- F1$selfcross(n = 1E3) # segregation population
RILS <- F1$dh(n = 1E3)      # inbreeds

## chainable application
F1$selfcross(n = 1, replace = TRUE)$selfcross(n = 1, replace = TRUE) # now it is F3!

## retrieving the genotypic data
Mrils <- genotype(RILS)                   # as numeric
Gid_wphase <- Gidj$genotype(phase = TRUE) # as character (phased)

## look for specific locus
## maybe useful for mutation or gene edit
F1$look('s1', phase = TRUE) # with phase
F1$look('s22')              # as numeric [-1/0/1]

## define a infinitesimal trait:
## all loci with the same additive and dominance values
infty <- set_infty(spp, m = 0, a = 1, d = .5) # partial dominance

## sampling some loci to define another trait (quantitative)
genes <- data.frame(snp = sample(ToyMap$snp, size = 10),
                    add = rnorm(10), dom = rnorm(10))

## define a quantitative trait
## given a table with additive and dominance values for each locus
quant <- set_quant(spp, m = 10, data = genes)

## evaluating breeding values regarding traits
## as a method of specimen
alphas_infty <- sapply(F2, function(x) infty$alpha(x))
hist(alphas_infty) ## distribution of the genotypic values

## as a method of trait
alphas_quant <- sapply(RILS, function(x) x$alpha(quant))
hist(alphas_quant) ## distribution of the genotypic values

###########################################################################
## user extensions for meiosis recombination

## define your own meiosis process
## "fixed" meiosis i.e., recombinations will occur at fixed points in the chromosome
FixedSrc <- '
// [[Rcpp::depends(isqg)]]
# include <isqg.h> // loading headers of the package

// fixed point recombination at position the middle of the chromosome
// Map is an alias for std::vector<double>
// Chromosome is a class with several slots/methods
// see inst/include/isqg for details
Map fixed(Chromosome * group) {

  double middle(group->get_length() / 2.) ;

  if (static_cast<bool>(R::rbinom(1.0, 0.5))) {
    return Map() ;
  } else {
    return Map(1, middle) ;
  }

}

// wrap the function as external pointer
// MPtr is a smart pointer
// FPtrM is a function pointer
// see inst/include/isqg/ for details
// [[Rcpp::export]]
MPtr fixedp() { return MPtr(new FPtrM(& fixed), true) ; }
'

## compile the extension
sourceCpp(code = FixedSrc, rebuild = TRUE)

## define a specie w/ custom meiosis
spp_custom <- set_specie(ToyMap, meiosis = fixedp())

## generate some gamete prototypes under your own meiosis
spp_custom$gamete(n = 20)

###########################################################################
## Comparing meiosis models

## fixed dimensions for the simulation
N <- 100 # n markers
n <- 1E4 # n gametes
L <- 2 # chromosome length in Morgans
C <- 2 # number of chromosomes

## test map with two chromosomes of length 2M with 100 markers evenly spaced
map <- data.frame(snp = paste0('s', 1:(N * 2)),
                  arrange(expand.grid(chr = 1:C, pos = seq(0, L, length.out = N)),
                          chr, pos),
                  stringsAsFactors = FALSE)

standard <- set_specie(map) # start isqg genome
gamS <- standard$gamete(n)  # sampling n gametes

## get gamete as numeric vector
breaks <- function(gam) {
    vec <- as.numeric(strsplit(gam, '')[[1]])
    return(vec)
}

## numeric gametes
cxS <- laply(gamS, breaks)

## fast way to get frequencies of recombination
src <- '
// [[Rcpp::plugins(cpp11)]]

# include <Rcpp.h>
# include <vector>
# include <algorithm>

// [[Rcpp::export]]
std::vector<int> freq2by2(std::vector<int> gamete) {

  int dim(gamete.size()) ;
  std::vector<int> out(std::pow(dim, 2), -1) ;

  for (auto row = 0; row < dim; row++)
    for (auto col = 0; col < dim; col++)
      out.at(row * dim + col) = gamete.at(row) != gamete.at(col) ;

  return out ;

}
'

## compiling auxiliary function
sourceCpp(code = src, rebuild = TRUE)

## loci by loci recombination frequency
frvS <- apply(apply(cxS, 1, freq2by2), 1, mean)

## digest
frdS <- mutate(merge(merge(cbind(expand.grid(snpi = map$snp, snpj = map$snp,
                                             stringsAsFactors = FALSE),
                                 fr = frvS),
                           map, by.x = 'snpi', by.y = 'snp'),
                     map, by.x = 'snpj', by.y = 'snp'),
               ## add a shift for markers in different chromosomes
               dist = ifelse(chr.x == chr.y, abs(pos.x - pos.y), L + .25))

## check against Haldane mapping function
## points represent the observed frequency of recombination as function of distance
## the red line is the expected frequency of recombination under Haldane function
ggplot(data = frdS, mapping = aes(x = dist, y = fr)) +
    geom_point(alpha = 1/10) +
    stat_function(mapping = aes(x = dist),
                  fun = function(x) (1 - exp(-2 * x)) / 2, # Haldane
                  colour = 'red')

## define independent segregation -- 2^k masks are equally probable
## all markers segregates independently from each other
Independent <- '
// [[Rcpp::depends(isqg)]]

# include <isqg.h> // loading headers of the package
# include <vector>
# include <algorithm>

// NOTE:
// independent markers
Map indep(Chromosome * group) {

  Map map(group->get_map()) ;

  for (auto it = 0; it < map.size(); it++)
    if (static_cast<bool>(R::rbinom(1., .5))) map.at(it) = 2. + 1. ;

  map.erase(std::remove(map.begin(), map.end(), 2. + 1.), map.end());

  return map ;

}

// wrap the function as external pointer
// [[Rcpp::export]]
MPtr indepp() { return MPtr(new FPtrM(& indep), true) ; }
'

## compile the extension for independent loci
sourceCpp(code = Independent, rebuild = TRUE)

custom <- set_specie(map, indepp()) # start isqg genome
gamC <- custom$gamete(n)  # sampling n gametes

## numeric gametes
cxC <- laply(gamC, breaks)

## loci by loci recombination frequency
frvC <- apply(apply(cxC, 1, freq2by2), 1, mean)

## digest
frdC <- mutate(merge(merge(cbind(expand.grid(snpi = map$snp, snpj = map$snp,
                                             stringsAsFactors = FALSE),
                                 fr = frvC),
                           map, by.x = 'snpi', by.y = 'snp'),
                     map, by.x = 'snpj', by.y = 'snp'),
               ## add a shift for markers in different chromosomes
               dist = ifelse(chr.x == chr.y, abs(pos.x - pos.y), L + .25))

## all two by two frequency of recombination are around 0.5 (independent)
## points represents observed frequency of recombination as a funtion of distance in M
ggplot(data = frdC, mapping = aes(x = dist, y = fr)) +
    geom_point(alpha = 1/10)

## recombination hotspots -- shifting beta random draws
## in this model, crossing over will preferentially occur at hotspots
## it is carried out by means of random draws from a beta distribution
Hotspots <- '
// [[Rcpp::depends(isqg)]]

# include <isqg.h> // loading headers of the package
# include <vector>
# include <algorithm>

// shifted beta distribution
// [[Rcpp::export]]
Rcpp::NumericVector shiftbeta(int n, double x) {

  return x * Rcpp::rbeta(n, 10.0, 2.5) ;

}

// NOTE:
// independent markers
Map hotspot(Chromosome * group) {

  double length(group->get_length()) ;

  int event(static_cast<int>(R::rpois(length))) ;

  if ( event == 0 ) {

    return Map() ;

  } else {

    Map chiasmata(Rcpp::as<Map>(shiftbeta(event, length))) ;

    std::sort(chiasmata.begin(), chiasmata.end()) ;

    return chiasmata ;

  }

}

// wrap the function as external pointer
// [[Rcpp::export]]
MPtr hotspotp() { return MPtr(new FPtrM(& hotspot), true) ; }
'

## compile the extension for recombination hotspots
sourceCpp(code = Hotspots, rebuild = TRUE)

hotspot <- set_specie(map, hotspotp()) # start isqg genome
gamH <- hotspot$gamete(n)  # sampling n gametes

## numeric gametes
cxH <- laply(gamH, breaks)

## get where crossing happen
cxX <- apply(cxH, 1, function(gam) which(as.logical(diff(gam))))

## digest
Counts <- data.frame(snp = map$snp,
                     chr = map$chr,
                     pos = map$pos,
                     count = hist(unlist(cxX),
                                  breaks = seq(.5, 200.5, 1),
                                  plot = FALSE)$counts,
                     stringsAsFactors = FALSE)

## distribution of crossing over across both chromosomes -- hotspots
## it is clear to see the hotspots of recombination
ggplot(data = subset(Counts, snp != 's100'), # dropping independent assortment!
       mapping = aes(x = pos, y = count)) +
    geom_bar(stat = 'identity') +
    facet_wrap(facets = ~ chr)

###########################################################################
## polymorphisms rather than diploid

## two bits together will have four alleles
droplets <- data.frame(snp = paste0('s', 2),   # two loci
                       chr = 1, pos = c(1, 1), # at the same position
                       stringsAsFactors = FALSE)

## initiate the droplet "specie"
XY <- set_specie(droplets)
XY$gamete(100) # there is no recombination between loci

## first individual has two alleles 11 and 12
i1 <- c('1 1', '1 2')
names(i1) <- droplets$snp
I1 <- import(XY, i1)

## second individual has the other two alleles 22 and 21
i2 <- c('2 2', '2 1')
names(i2) <- droplets$snp
I2 <- import(XY, i2)

## crossing then will have the four alleles segregating with equal proportion
CD <- t(genotype(cross(n = 1000, I1, I2), phase = TRUE))

## checking proportions
table(c(sapply(c(1, 3), # getting DNA tape **bear in mind the space**
               function(t) apply(CD, 1,
                                 function(g) paste(sapply(g, substring,
                                                          first = t,
                                                          last = t),
                                                   collapse = '')))))

## three bits together will have eight alleles
triplets <- data.frame(snp = paste0('s', 3),   # two loci
                       chr = 1, pos = c(1, 1, 1), # at the same position
                       stringsAsFactors = FALSE)

## initiate the triplet "specie"
XYZ <- set_specie(triplets)
XYZ$gamete(100) # there is no recombination between loci

## first individual has two alleles 111 and 112
i4 <- c('1 1', '1 1', '1 2')
names(i4) <- triplets$snp
I4 <- import(XYZ, i4)

## second individual has two alleles 121 122
i5 <- c('1 1', '2 2', '1 2')
names(i5) <- triplets$snp
I5 <- import(XYZ, i5)

## third individual has two alleles 211 212
i6 <- c('2 2', '1 1', '1 2')
names(i6) <- triplets$snp
I6 <- import(XYZ, i6)

## forth individual has two alleles 221 222
i7 <- c('2 2', '2 2', '1 2')
names(i7) <- triplets$snp
I7 <- import(XYZ, i7)

## putting _in silico_ genotypes within a list
pop <- list(I4, I5, I6, I7)

## function to simulate random crosses between all four individuals
random <- function(w) {
    pair <- sample(1:4, 2, replace = FALSE)
    x <- pop[[pair[1]]]
    y <- pop[[pair[2]]]
    z <- cross(n = 1, x, y)
    return(z)
}

## crossing then will have the four alleles segregating with equal proportions
CT <- t(genotype(sapply(1:1000, random), phase = TRUE))

## checking proportions
table(c(sapply(c(1, 3), # getting DNA tape **bear in mind the space**
               function(t) apply(CT, 1,
                                 function(g) paste(sapply(g, substring,
                                                          first = t,
                                                          last = t),
                                                   collapse = '')))))

## \EOF
###########################################################################
